/* Solution (part) of Exercise 4 - Numerics of Transport in Porous Media

   This implements the class providing the boundary conditions. It
   reads the configure file format as given on the exercise sheet.

   This implementation makes use of the LGPL library libconfig which
   has to be available in order to compile this header file.

   Furthermore it requires implementations of the classes implemented
   in exercise 1 and 2 assumed within the header files
   linear_algebra.hh and grid.hh .
 */

#ifndef BOUNDARY_COND_HH
#define BOUNDARY_COND_HH

#include "Vector.hh"
#include "DenseMatrix.hh"
#include "grid.hh"
#include <libconfig.h++>


using namespace std;
using namespace libconfig;

class Sources{
public:
  virtual double operator()(const Vector & ) const
  { return 0; }

  virtual ~Sources(){}
};




class BoundaryConditions{
public:

protected:
  struct Condition{
    vector<Vector> ranges;
    double value;
    bool is_dirichlet;
    bool is_gravity;
  };

private:
  Vector extent_;
  const static double epsilon = 1e-10;

  vector<Condition> neumann_conditions;
  vector<Condition> dirichlet_conditions;
  Condition default_condition;
  bool default_condition_specified;

  // Extracts the floating type values from a range given in a string
  // like "0.1..1.3". The lower value is reduced by epsilon and the
  // higher is increased by epsilon.
  Vector split_range(string range)
  {
    Vector v_range(2);
    const size_t pos = range.find("..");
    if(pos != string::npos){
      v_range[1] = atof(range.substr(pos+2,range.length()-pos-2).c_str());
      v_range[0] = atof(range.substr(0,pos).c_str());
    }
    else{
      v_range[1] = atof(range.c_str());
      v_range[0] = atof(range.c_str());
    }
    cout << v_range[0] << "-eps .. " << v_range[1] << "+eps";
    v_range[0]-= epsilon;
    v_range[1]+= epsilon;

    if(!isfinite(v_range[0]) || !isfinite(v_range[1])){
      cerr << "Error: Got a bad range argument \"" << range <<"\"." << endl;
      exit(1);
    }
    return v_range;
  }

  void read_condition_list(Setting & conditions, bool is_default)
  {
    const int n_conditions = conditions.getLength();
    if(n_conditions > 0){
      if(is_default)
	cout << endl << "Found default boundary condition:" << endl;
      else
	cout << endl << "Found spatially defined boundary conditions:" << endl;

      for(int i=0; i<n_conditions; ++i){
	if(
	   !conditions[i].exists("type") ||
	   (!conditions[i].exists("xrange") && !is_default) ||
	   (!conditions[i].exists("yrange") && !is_default) ||
	   !conditions[i].exists("value") ||
	   (Grid::dimensions == 3 && !conditions[i].exists("zrange") && !is_default)
	   ){
	  cerr << "Error: Boundary condition nr "
	       <<i<<" was not well defined."<< endl;
	  exit(1);
	}
	Condition c;
	cout << endl << "Boundary condition nr "<<i<<":" << endl;
	if(!is_default){
	  cout << "Found x range to be ";
	  c.ranges.push_back(split_range(conditions[i]["xrange"]));
	  cout << endl << "Found y range to be ";
	  c.ranges.push_back(split_range(conditions[i]["yrange"]));
	  cout << endl;
	  if(Grid::dimensions==3){
	    cout  << "Found z range to be ";
	    c.ranges.push_back(split_range(conditions[i]["zrange"]));
	    cout << endl;
	  }
	}
	if(conditions[i]["value"].isNumber())
	  c.value = conditions[i]["value"];
	else{
	  string s = conditions[i]["value"];
	  c.value = atof(s.c_str());
	}
	if(!isfinite(c.value)){
	  cerr << "Error: Bad value in boundary condition nr " << i << endl;
	  exit(1);
	}
	string bc_type = conditions[i]["type"];
	cout << "Type is : "<< bc_type << endl;

	if(is_default)
	  default_condition_specified = true;
	    
	if(bc_type.find("Neumann")!=string::npos){
	  cout << "Value is : " << c.value << endl;

	  c.is_dirichlet = false;
	  if(is_default)
	    default_condition = c;
	  else
	    neumann_conditions.push_back(c);
	}
	else if(bc_type.find("GravityDirichlet")!=string::npos){
	  cout << "Value is : " << c.value;
	  c.is_dirichlet = true;
	  c.is_gravity = true;
	  if(is_default)
	    default_condition = c;
	  else
	    dirichlet_conditions.push_back(c);
	}
	else if(bc_type.find("Dirichlet")!=string::npos){
	  cout << "Value is : " << c.value;
	  c.is_dirichlet = true;
	  c.is_gravity = false;
	  if(is_default)
	    default_condition = c;
	  else
	    dirichlet_conditions.push_back(c);
	}
	else{
	  cerr << "Error: Bad type in boundary condition nr " << i<< endl;
	  exit(1);
	}
      } // i -> n_conditions
    } // if n_conditions > 0
  }

  const Condition * isInside(const Vector & position, const vector<Condition> conditions) const
  {
    typedef vector<Condition>::const_iterator const_iterator;
    const_iterator it = conditions.begin();
    const_iterator eit = conditions.end();

    for(; it!=eit; ++it){
      bool inside=true;
      for(int d=0; d<Grid::dimensions; ++d){
	if(position[d] < it->ranges[d][0] || position[d] > it->ranges[d][1]){
	  inside=false;
	  break;
	}
      }
      
      if(inside){
	return &(*it);
      }
    }
    return 0;
  }

  void init(const string filename)
  {
    Config cfg;
    try{
      cfg.readFile(filename.c_str());
      cfg.setAutoConvert(true);
    
      if(cfg.exists("boundary_conditions.Spatial")){
	
	Setting & conditions = cfg.lookup("boundary_conditions.Spatial");
	read_condition_list(conditions,false);
      } // if spatial conditions define
      
      if(cfg.exists("boundary_conditions.Default")){
	Setting & conditions = cfg.lookup("boundary_conditions.Default");
	read_condition_list(conditions,true);
	}

      if(!default_condition_specified)
	cout << "No default boundary condition specified. "
	  "Assuming homogeneous Dirichlet conditions." << endl;
    } // try
    catch(ParseException & exc){
      cerr << "Error reading config file "<< filename << ": " << endl;
      cerr << "\t " << exc.getError() <<" in line "<< exc.getLine() << endl;
      exit(1);
    }
    catch(FileIOException & exc){
      cerr << "Error reading config file "<< filename <<  endl;
      cerr << "\t File could not be read at all." <<  endl;
      exit(1);
    }
    catch(SettingTypeException & exc){
      cerr << "Error: One of your boundary conditions settings was of bad type." << endl;
      exit(1);
    }
  }

public:

  //! Constructor. Reads boundary conditions from file
  BoundaryConditions(const Grid & grid, const string filename)
    : extent_(grid.extent()), default_condition_specified(false)
  {
    init(filename);
  }

  template<class D>
  BoundaryConditions(const Grid & grid, const string filename, const D &)
    : extent_(grid.extent()), default_condition_specified(false)
  {
    init(filename);
  }


  //! Checks whether dirichlet conditions were specified at
  //! coordinates given by position. If value pointer is valid, the
  //! dirichlet value is stored in its destination if function returns
  //! true.
  bool is_dirichlet(const Vector & position, double * value = 0) const
  {

    if(const Condition* c = isInside(position,dirichlet_conditions)){
      if(value){
	if(c->is_gravity)
	  *value = (c->value - position[GRAVITY_AXES]) * RHO * GRAVITY;
	else
	  *value = c->value;
      }
      return true;
    }

    if(isInside(position,neumann_conditions)){
      return false;
    }
    
    if(default_condition_specified){
      if(default_condition.is_dirichlet){
	if(value){
	  if(default_condition.is_gravity)
	    *value = (default_condition.value - position[GRAVITY_AXES]) * RHO * GRAVITY;
	  else
	    *value = default_condition.value;
	}
	return true;
      }
    }
    else{// homogeneous dirichlet assumed
      if(value){
	    *value = 0;
      }
      return true;
    }
    
    return false;
  }

  //! Returns the Dirichlet value at the boundary coordinates given by
  //! position. The return value is not well defined if the
  //! coordinates point to a Neumann boundary!
  double get_dirichlet(const Vector & position) const
  {
    if(const Condition* c = isInside(position,dirichlet_conditions)){
	if(c->is_gravity)
	  return (c->value - position[GRAVITY_AXES]) * RHO * GRAVITY;
	else
	  return c->value;
    }

    if(default_condition_specified){
      if(default_condition.is_dirichlet){
	if(default_condition.is_gravity)
	  return  (default_condition.value - position[GRAVITY_AXES]) * RHO * GRAVITY;
	else
	  return  default_condition.value;
      }
      else{
	assert(false);
      }
    }
    else{// homogeneous dirichlet assumed

      // Notice! This branch may be reached if position points to a spatial
      // Neumann boundary.
	return 0;
    }

    assert(false);
    return 0;
  }

  //! Returns the Neumann value at the boundary coordinates given by
  //! position.
  double get_neumann(const Vector & position) const
  {
    if(const Condition* c = isInside(position,neumann_conditions)){
	return c->value;
    }

    if(default_condition_specified){
      if(!default_condition.is_dirichlet)
	  return default_condition.value;
      else{
	assert(false);
      }
    }

    assert(false);
    return 0;
  }

};

#endif
